# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.dianomic.com
# FOGLAMP_END

"""
Moving Measure Calculation
It performs the following
    - provides the choices in weightage methodology
    - provides the choices in measure
    - calculate the moving measure
"""


__author__ = "Mohit Rajput"
__copyright__ = "Copyright (c) 2023 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


from foglamp.common import logger
_LOGGER = logger.setup(__name__)

try:
    import numpy as np
    from typing import Union
    from collections import deque
    from abc import ABC, abstractmethod

    _LOGGER.debug('Python packages [1/2] in "{}" loaded successfully'.format(__file__))
except ImportError as e:
    _LOGGER.error(
        'Failed to load Python packages in "{}". {}'.format(
            __file__, str(e)
        )
    )

try:
    from scipy.stats import sem, skew, kurtosis
    _LOGGER.debug('Python packages [2/2] in "{}" loaded successfully'.format(__file__))
except ImportError as e:
    _LOGGER.error(
        'Failed to load "scipy" Python package in "{}". {}'.format(
            __file__, str(e)
        )
    )


class MovingMeasure(ABC):
    """ Calculate Moving Measure

    Reference:
        https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.ewm.html
        https://pandas.pydata.org/docs/user_guide/window.html#window-exponentially-weighted
        https://docs.scipy.org/doc/scipy/reference/signal.windows.html#module-scipy.signal.windows
    """

    @abstractmethod
    def calc_measure(self):
        pass


class SMM(MovingMeasure):
    """ Simple (Non-Weighted or weight = 1) Moving Measure Calculation Class.
    """

    measures = {
        "mean": np.mean,
        "median": np.median,
        "min": np.min,
        "max": np.max,
        "std": np.std,
        "sum": np.sum,
        "var": np.var,
    }
    try:
        measures["sem"] = sem
        measures["skew"] = skew
        measures["kurtosis"] = kurtosis
    except Exception as e:
        pass

    __slots__ = ["_queue", "measure_func"]

    def __init__(self, window_size: int = 10, measure: str = "mean"):
        """ Simple Moving Measure.

        Args:
            window size:
                rolling window size
                defines the amount of data that will be kept for moving-measure calculation
            measure: aka the measure that is to be calculated
                options:
                    'mean'   : average
                    'median' : median
                    'min'    : minimum
                    'max'    : maximum
                    'std'    : standard deviation
                    'sem'    : standard error of the mean
                    'sum'    : sum
                    'var'    : unbiased variance
                    'skew'   : unbiased skew
                    'kurtosis': unbiased kurtosis
        """
        self._queue = deque(maxlen=window_size)
        self.measure_func = self.measures[measure.lower()]

    @property
    def allowed_measures(self):
        """ Return the allowed measures.
        """
        return list(self.measures.keys())

    def calc_measure(self, curr_value: Union[int, float]):
        """ Calculate Simple Moving Measure.

        Args:
            curr_value: Get the new reading for the series

        Return:
            the latest calculated value
        """
        self._queue.append(curr_value)
        moving_measure = self.measure_func(self._queue)
        return moving_measure


class EWMA(MovingMeasure):
    """ Exponentially Weighted Moving Average Calculation Class.
    """

    __slots__ = ["_prev_ema", "alpha"]

    def __init__(
        self,
        alpha: Union[int, float, None] = 0.2,
        span: Union[int, float, None] = None,
        com: Union[int, float, None] = None,
        halflife: Union[int, float, None] = None,
    ):
        """
        Args:
            ONLY ONE OF THESE SHOULD BE NOT NONE
            alpha:
                the smoothing factor
                default: 0.2
            span :
                It corresponds to what is commonly called an “N-day EW moving average”.
                default: None
                alpha = 2 / (span+1) for span >= 1
            com :(Center of mass)
                It has a more physical interpretation and can be thought of in terms of span
                default: None
                alpha = 1 / (1+c) for com >= 0
            halflife:
                It is the period of time for the exponential weight to reduce to one half
                default: None
                alpha = 1 - exp**(log 0.5 / halflife) for halflife > 0
        """
        self._prev_ema = None
        self.alpha = EWMA.check_and_get_ema_alpha(alpha, span, com, halflife)

    @staticmethod
    def check_and_get_ema_alpha(
        alpha: Union[int, float, None] = None,
        span: Union[int, float, None] = None,
        com: Union[int, float, None] = None,
        halflife: Union[int, float, None] = None,
    ):
        """ Single value within the arg should not be None.
        """
        # check if only a single measure is provided
        check_val = sum([alpha is None, span is None, com is None, halflife is None])
        if check_val != 3:
            raise Exception(
                'Only a single value among "alpha", "com", "span", "halflife"'
                + " in EMA should be not None."
                + 'Currently: "alpha {}", "com {}", "span {}", "halflife {}"'.format(
                    alpha, com, span, halflife
                )
            )

        # alpha calculation from span
        if span is not None:
            if span >= 1:
                alpha = 2 / (span + 1)
            else:
                raise Exception(
                    '"span" value should be greater than equal to 1'
                )

        # alpha calculation from com
        if com is not None:
            if com >= 0:
                alpha = 1 / (1 + com)
            else:
                raise Exception(
                    '"center of mass" value should be greater than equal to 0'
                )

        # alpha calculation from halflife
        if halflife is not None:
            if halflife > 0:
                alpha = 1 - np.exp ** (np.log(0.5) / halflife)
            else:
                raise Exception("Halflife value should be greater than 0")

        return alpha

    def calc_measure(self, curr_value: Union[int, float]):
        """ Get the new reading and calculate the Moving-Measures.

        In Exponential weighted window each prior point is being exponentially weighted down
        relative to the current point.
        At a time, any single value in [] should have some value

        Args:
            curr_value: Get the new reading for the series

        Return:
            EMA(t) = alpha * val(t) + (1-alpha) * EMA(t-1)
        """
        # Calculate the exponent measure
        if self._prev_ema is None:
            _curr_ema = curr_value
        else:
            _curr_ema = (
                self.alpha * curr_value + (1 - self.alpha) * self._prev_ema
            )

        # update the value
        self._prev_ema = _curr_ema

        return _curr_ema


class LWMA(MovingMeasure):
    """Linearly Weighted Moving Average Calculation Class."""

    __slots__ = ["_queue"]

    def __init__(self, window_size: int = 10):
        self._queue = deque(maxlen=window_size)

    def _calc_wm_nsum(self, n: int):
        return n * (n + 1) / 2

    def calc_measure(self, curr_value: Union[int, float]):
        """ Calculate Linear Weighted Moving Average.

        Args:
            curr_value: Get the new reading for the series
        Return:
            WMA(n) = ( n*curr_value + WMA(n-1) * ((n-1)*n)/2 ) /  (n*(n+1)/2)
        """
        # appending the new-value into it's series
        self._queue.append(curr_value)

        # get the window size (aka n) - In the starting we won't have that many
        # datapoint hence designed this way
        curr_n = len(self._queue)

        # Create Linear Weights
        wts = np.arange(0, curr_n)

        # Calculating the moving-measure - alter num = sum(np.array(_queue)*wts)
        _curr_wma = np.dot(np.array(self._queue), wts) / self._calc_wm_nsum(
            curr_n
        )
        return _curr_wma


class MovingMeasureExec:
    """ Execute the Moving Measure with FogLAMP.
    """
    __slots__ = ["moving_measure", "mm_cfg"]

    @staticmethod
    def parse_config(config):
        """ Configuration Parser
        """
        # Inputs
        inputs = {
            "asset": config["in_asset"]["value"],
            "datapoint": config["in_datapoint"]["value"],
        }

        # Aggregating params
        weight_strategy = {
            "Simple (Non-Weighted)": "swm",
            "Linearly Weighted": "lwm",
            "Exponential Weighted": "ewm",
        }
        params = {
            # Calc-Config
            "wt_strat": weight_strategy[config["cc_wt_strategy"]["value"]],
            "window_size": int(config["cc_window_size"]["value"]),
            # Simple weighted Moving Measure (non-weighted or weighted 1)
            "swm": {
                "measure": config["cc_s_measure"]["value"].lower(),
            },
            # Weighted Moving Measure
            "lwm": {},
            # Exponential Moving Measure
            "ewm": {
                "parameter": config["cc_ema_parameter"]["value"].lower(),
                "alpha": float(config["cc_ema_alpha"]["value"]),
                "span": float(config["cc_ema_span"]["value"]),
                "com": float(config["cc_ema_com"]["value"]),
                "halflife": float(config["cc_ema_halflife"]["value"]),
            }
        }

        # Value Conversion - converting default value to None w.r.t. parameter
        if params["ewm"]["parameter"] != "alpha":
            params["ewm"]["alpha"] = None
        if params["ewm"]["parameter"] != "span":
            params["ewm"]["span"] = None
        if params["ewm"]["parameter"] != "center of mass":
            params["ewm"]["com"] = None
        if params["ewm"]["parameter"] != "halflife":
            params["ewm"]["halflife"] = None

        # Output
        auto_suffix = config["out_datapoint_suffix_auto"]["value"] == "true"
        out_suffix = config["out_datapoint_suffix"]["value"]

        # create output suffix automatically
        if auto_suffix:
            out_suffix = '_{}{}_{}'.format(
                params["wt_strat"],
                '_{}'.format(params["swm"]["measure"]) if params["wt_strat"] == 'swm' else 'a',
                'al{}'.format(
                    round(EWMA.check_and_get_ema_alpha(**{
                        k: it
                        for k, it in params["ewm"].items()
                        if k in ["alpha", "span", "com", "halflife"]
                    }), 3)
                ).replace('.', 'd') if params["wt_strat"] == 'ewm' else params["window_size"]
            )

        # Enable
        enable = config["enable"]["value"] == "true"

        # --------------- << Creating New Config >> --------------- #
        new_config = {}
        new_config["enable"] = enable
        new_config["input"] = inputs
        new_config["params"] = params
        new_config["output"] = {"datapoint": (inputs["datapoint"] + out_suffix)}

        return new_config

    def __init__(self, cfg):

        self.moving_measure = None
        self.mm_cfg = MovingMeasureExec.parse_config(cfg)

        strategy = self.mm_cfg["params"]["wt_strat"]
        if strategy == "swm":
            self.moving_measure = SMM(
                self.mm_cfg["params"]["window_size"],
                self.mm_cfg["params"]["swm"]["measure"],
            )
        elif strategy == "lwm":
            self.moving_measure = LWMA(self.mm_cfg["params"]["window_size"])
        elif strategy == "ewm":
            self.moving_measure = EWMA(
                **{
                    k: it
                    for k, it in self.mm_cfg["params"]["ewm"].items()
                    if k in ["alpha", "span", "com", "halflife"]
                }
            )
        else:
            raise Exception(
                '{}: "{}" is not supported'.format(
                    cfg["cc_wt_strategy"]["displayName"],
                    cfg["cc_wt_strategy"]["value"],
                )
            )

    def calc_and_add_reading(self, data: Union[list, dict]):
        """ Calculate & appends Measure.

        Args:
            data:
                eg. [
                        {'readings': {'sinusoid': -0.97}, 'asset': 'sin_asset', 'id': 0, 'ts': '2023-07-06 12:21:44.663815+00:00', 'user_ts': '2023-07-06 12:21:44.663815+00:00'},
                        {'readings': {'sinusoid': -0.99}, 'asset': 'sin_asset', 'id': 0, 'ts': '2023-07-06 12:21:45.663833+00:00', 'user_ts': '2023-07-06 12:21:45.663833+00:00'},
                        {'readings': {'sinusoid': -1.0}, 'asset': 'sin_asset', 'id': 0, 'ts': '2023-07-06 12:21:46.664387+00:00', 'user_ts': '2023-07-06 12:21:46.664387+00:00'},
                        {'readings': {'sinusoid': -0.99}, 'asset': 'sin_asset', 'id': 0, 'ts': '2023-07-06 12:21:47.663813+00:00', 'user_ts': '2023-07-06 12:21:47.663813+00:00'}
                    ]
        """
        # Check if filter enabled or disabled
        if not self.mm_cfg["enable"]:
            return

        # Type conversion
        if isinstance(data, dict):
            data = [data]

        _msg = "A total of {} elements received in the data list.".format(len(data))
        _LOGGER.debug(_msg)

        # Rotating the data.
        # Note: reference are to be maintained oterwise the added data
        # will be lost;
        for d in data:
            # Focus on specific data that is of interest
            if (self.mm_cfg["input"]["asset"] == d["asset"]) and (
                self.mm_cfg["input"]["datapoint"] in d["readings"]
            ):
                calc_value = None
                # get current value & calculate measure
                try:
                    # change dtype
                    curr_value = float(d["readings"][self.mm_cfg["input"]["datapoint"]])
                    calc_value = self.moving_measure.calc_measure(curr_value)
                except Exception as e:
                    raise Exception(
                        'Unexpected error within "moving_measure" calculation. {}'.format(
                            str(e)
                        )
                    )

                # Adding value to readings
                if calc_value:
                    d["readings"][self.mm_cfg["output"]["datapoint"]] = calc_value
            else:
                # keeping to debug as this might be expected too
                _msg = '[Note] within this reading either input_asset("{}") or input_datapoint("{}") not found.'.format(
                    self.mm_cfg["input"]["asset"] == d["asset"],
                    self.mm_cfg["input"]["datapoint"],
                )
                _LOGGER.debug(_msg)

        return data
